//TextNavigate.cpp

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "plugin.hpp"

#define FarSprintf FSF.sprintf

#include "crtdll.h"
#include "mem.hpp"
#include "CRegExp.hpp"
#include "Dialog.hpp"
#include "sgml.hpp"
#include "pclasses.hpp"
#include "TextNavigate.hpp"
#include "TextNavigate-lng.h"

static struct PluginStartupInfo Info;
static struct FarStandardFunctions FSF;
PTextNavigate TextNavigate;

#define MAX_CHAR_SET_LENGTH 30

static char s_AdditionalLetters[ MAX_CHAR_SET_LENGTH ];
static int active;
static int adddigits;
static int casesensitive;
static int cyclicsearch;
static int searchselection;

static struct _REG_STR
{
  const char *sss;
  const char *active;
  const char *adddigits;
  const char *casesensitive;
  const char *cyclicsearch;
  const char *searchselection;
  const char *AdditionalLetters;
} reg_key = {

  "%s%s%s",
  "Active",
  "ConsiderDigitAsChar",
  "CaseSensitive",
  "CyclicSearch",
  "SearchSelection",
  "AdditionalLetters",
};

inline const char *get_msg(int id)
{
  return( Info.GetMsg(Info.ModuleNumber,id) );
}

#include "mem.cpp"
#include "reg.c"
#include "CRegExp.cpp"
#include "tools.cpp"
#include "Dialog.cpp"
#include "sgml.cpp"
#include "pclasses.cpp"

#ifdef __cplusplus
extern "C"
#endif

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *PSInfo)
{
  ::Info = *PSInfo;
  FSF = *PSInfo->FSF;
  Info.FSF = &FSF;
#ifdef _DEBUG
  CREATE_LOG_FILE();
#endif
#ifdef _DEBUG
  char tmp_str[500];
  FarSprintf(tmp_str, "CSearchPaths=%d, CMethod=%d, CClass=%d,\r\n \
CLanguage=%d, CXMLFilesColl=%d,\r\n \
CXMLFile=%d;CTextNavigate=%d, ENV_VARS=%d, _FND=%d",
     sizeof(CSearchPaths), sizeof(SMethod), sizeof(SClass),
     sizeof(SLanguage), sizeof(CXMLFilesColl),
     //sizeof(CLangColl),
     sizeof(CXMLFile), sizeof(CTextNavigate),
     sizeof(ENV_VARS), sizeof(_FOUND) );

  WRITE_TO_LOG_FILE(tmp_str);
#endif
  FarSprintf(PluginRegRootKey, "%s\\%s", Info.RootKey, "Text Navigate");
  GetRegKey(HROOT, "", reg_key.active, active, 1);
  GetRegKey(HROOT, "", reg_key.adddigits, adddigits, 1);
  GetRegKey(HROOT, "", reg_key.casesensitive, casesensitive, 1);
  GetRegKey(HROOT, "", reg_key.cyclicsearch, cyclicsearch, 1);
  GetRegKey(HROOT, "", reg_key.searchselection, searchselection, 0);
  GetRegKey(HROOT, "", reg_key.AdditionalLetters, s_AdditionalLetters, "---_", MAX_CHAR_SET_LENGTH);

  TextNavigate = new CTextNavigate();
}//SetStartupInfo

void WINAPI _export GetPluginInfo(struct PluginInfo *PInfo)
{
  PInfo->StructSize = sizeof(struct PluginInfo);
  PInfo->Flags = PF_DISABLEPANELS|PF_EDITOR;

  static const char *PMStrings[] = { get_msg(STitle) };

  PInfo->PluginMenuStrings = PMStrings;
  PInfo->PluginMenuStringsNumber = 1;
  PInfo->PluginConfigStrings = PMStrings;
  PInfo->PluginConfigStringsNumber = 1;
}//GetPluginInfo

int WINAPI _export GetMinFarVersion(void){ return( MAKEFARVERSION(1,70,0) ); };

HANDLE WINAPI _export OpenPlugin(int, int)
{
  if( TextNavigate )
    TextNavigate->ShowPluginMenu();

  return( INVALID_HANDLE_VALUE );
}//OpenPlugin

void  WINAPI _export ExitFAR()
{
  delete TextNavigate;
}//ExitFAR

int WINAPI _export Configure(int)
{
  if( TextNavigate )
    TextNavigate->config_plugin();
  return( FALSE );
}//Configure

int WINAPI _export ProcessEditorInput(const INPUT_RECORD *Rec)
{
  if( !active || !TextNavigate )return( FALSE );
  int res = TextNavigate->ProcessEditorInput(Rec);
  return( res );
}//ProcessEditorInput

int WINAPI _export ProcessEditorEvent(int Event, void *Param)
{
  if( !active )return( 0 );
  int res = TextNavigate->ProcessEditorEvent(Event, Param);
  return( res );
}//ProcessEditorEvent


