//pclasses.hpp

#ifndef _PCLASSES_H_
#define _PCLASSES_H_

typedef struct _FOUND
{
  unsigned int path_index;
  const char *file_name;
  const char *file_path;
  //DWORD FileSize;
} FOUND, _FND, *LPFND;

#define MAX_ENV_VARS 16
typedef struct ENV_VAR{
 char *EnvVar;
 char *EnvVarValue;
} ENV_VARS[ MAX_ENV_VARS ];

#define PWL 0x400
#define PWC 0x100      //- pathways
#define WFC 50         //- wasfound
typedef char *pw_array[ PWC ];
typedef _FND wf_array[ WFC ]; // 

typedef struct SLanguage *PLanguage;
typedef class CSearchPaths *PSearchPaths;

class CSearchPaths
{
  char *s_PathWays; //[ PWL ];
  int s_PathWays_len;
  char *file_name; //[ MAX_PATH ];
  //char *pathways[ PWC ]; // 
  pw_array *pathways;
  char *last_path; //[ MAX_PATH ];
  char *find_file_name; //  
  char *find_file_path; //  
  UINT pw_count; // 
  ENV_VARS env_vars;
  wf_array *was_found;
  UINT wf_count; //  
  int EnvVarsCount;
  //char def_path[ MAX_PATH ];

  char *get_def_path();
  void ResolveEnvVars();
  int AlreadyFound(char *path, char *file_name);
  int get_full_file_name(char *dest, LPFND fnd); //!!!

  void free_find_data();
  void PrepareForSearch();
  void free_pathways();
  void MakePathWays();
  void FindAllSubDirs(char *RootDir);
 public:
  CSearchPaths(PSgmlEl elem);
  ~CSearchPaths();

  bool ProcessCtrlEnter(PLanguage Language);

  int get_filename();
  int find_file(char *ExcludedFileExts);
  int get_full_file_name(char *dest, int pos);
  void ClearAfterSearch();

  void AddSearchPaths(PSgmlEl elem);
  int ShowSelectFileMenu();
};

typedef struct SMethod *PMethod;
struct SMethod
{
  SMethod(PSgmlEl elem, char *Tag);
  ~SMethod();

  char *Name, *Type, *Definition, *Implementation;
};

typedef struct SClass *PClass;
struct SClass
{
  SClass(PSgmlEl elem);
  ~SClass();

  char *Name, *Definition, *End;
  PMethod Method;
};

typedef class CXMLFile *PXMLFile;

struct SLanguage
{
  SLanguage(PXMLFile owner, PSgmlEl elem = NULL);
  ~SLanguage();

  char *name, *FileExts, *ExcludedFileExts, *SourceFiles, *Headers; //, *Desc;
  PClass Class;
  
  PMethod Method;
  PXMLFile Owner;
};

//typedef class CLangColl *PLangColl;

class CXMLFile
{
  char *XMLFileName;
  char *filepath; //[ MAX_PATH ];

  PSgmlEl searchbasetag(PSgmlEl base);
  bool loaddata(const char *data, int len);
  bool includefile(char *param);
  void readcdata(PSgmlEl basetag);
 public:
  CXMLFile();
  CXMLFile(char *xml_file_name);
  ~CXMLFile();
  void init(char *file_name);
  void AddSearchPaths(PSgmlEl elem);

  //PLangColl Languages;
  PLanguage Language;
  PSearchPaths SearchPaths;
};

#pragma warn -inl
template <class T> class CRefColl
{
  int id;
  bool main;
 public:
  T *Ref;
  CRefColl *next, *prev;

  CRefColl(bool Main = FALSE, T *ref = NULL)
  {
    id = -1;
    main = Main;
    next = prev = this;
    Ref = ref;
  }
  ~CRefColl()
  {
    if( !main )return;
    CRefColl *fl = next;
    while( fl != this )
    {
      fl = fl->next;
      delete fl->prev->Ref;
      delete fl->prev;
    }
    delete Ref;
  }

  CRefColl *GetRefColl(int fid)
  {
    CRefColl *fl = next;
    while( fl != this )
    {
      if( fl->id == fid ) return fl;
      fl = fl->next;
    }
    return NULL;
  }

  T *Get(int fid)
  {
    CRefColl *fl = next;
    while( fl != this )
    {
      if( fl->id == fid ) return fl->Ref; //fl;
      fl = fl->next;
    }
    return NULL;
  }

  T *Add(int fid = -1, T *ref = NULL)
  {
    CRefColl *fl = prev;
    fl->next = new CRefColl(false);
    fl->next->prev = fl;
    fl = fl->next;
    fl->next = this;
    prev = fl;
    fl->id = fid;
    fl->Ref = ref;
    return ref;
  }
  bool Delete(int fid)
  {
    CRefColl *fl = GetRefColl(fid);
    if( !fl )return( FALSE );

    fl->prev->next = fl->next;
    fl->next->prev = fl->prev;
    if( fl->Ref )delete fl->Ref;
    delete fl;
    return TRUE;
  }
};
/*
class CLangColl: public CRefColl<SLanguage>
{
 public:
  CLangColl(PLanguage Language = NULL)
    : CRefColl<SLanguage>(TRUE, Language) { }
};
*/
typedef class CXMLFilesColl *PXMLFilesColl;
class CXMLFilesColl: public CRefColl<CXMLFile>
{
 public:
  CXMLFilesColl(PXMLFile XMLFile = NULL)
    : CRefColl<CXMLFile>(TRUE, XMLFile) { }
};

#define MAX_WORD_LENGTH 100

typedef class CTextNavigate *PTextNavigate;
class CTextNavigate
{
  CRegExp reg;
  int FKeyCode, FPrevKeyCode;

  PXMLFile XMLFile;
  PXMLFilesColl XMLFilesColl;

  char word[ MAX_WORD_LENGTH ];

  char MethodType[80];
  char MethodName[100];
  char ClassName[80];
  char MethodDefinition[100];
  PClass Class; PMethod Method;

  int strreplace(char *str, const char *pattern, const char *value);
  void ReplaceSpecRegSymbols(char *str);
  int GetMatch(char *Match, const SMatches &m, const char *str, int n);

  bool SearchBackward(char *RegExpr, int &CurLine, const char *&StringText, SMatches &m);
  bool SearchForward(char *RegExpr, int &CurLine, const char *&StringText, SMatches &m);
  bool IsInClassDefinition(int CurLine);
  bool IsGlobalMethodImplementation(int CurLine);
  int SearchForMethodImplementation(int &CurLine, int GlobalProc);
  bool SearchForMethodImplementation2(int &CurLine, int &x_pos);
  bool SearchForMethodDefinition2(int &CurLine, bool GlobalProc);
  bool SearchForMethodDefinition(int &CurLine, int &x_pos, bool GlobalProc);
  bool FindMethodDefinition(int &CurLine, int &x_pos);
  bool FindMethodImplementation(int &CurLine, int &x_pos);

  int FindHeaderOrSourceFileName();
  int FindInAnotherFile(bool Header);
  //void Colorize(int strlen_word);
 public:
  CTextNavigate();
  ~CTextNavigate();

  int ProcessEditorInput(const INPUT_RECORD *Rec);
  int ProcessEditorEvent(int Event, void *Param);

  int processCtrlAltUpDown(int FKeyCode, int FPrevKeyCode);
  int processCtrlEnter(void);
  int processCtrlShiftUpDown(int FKeyCode);
  int OpenHeaderOrSource(void);

  int ShowPluginMenu();
  void config_plugin();
};

typedef struct FFarMenuItem
{
  char *Text;
  int  Selected;
  int  Checked;
  int  Separator;
} *PFFarMenuItem;

typedef struct FarMenuItem *PFarMenuItem;
typedef class CPluginMenu *PPluginMenu;

class CPluginMenu: public RefArray<FarMenuItem>
{
 public:
  //CPluginMenu( int NumElemens );
  CPluginMenu( int NumElements, PFFarMenuItem Elements, int FNumElements = -1 );
  FarMenuItem *Setup( int num, const FFarMenuItem &mi );
  void Setup( int From, PFFarMenuItem p, int count );
  int Execute(const char *Title, const char *Bottom);
};


#endif