//tools.cpp

SCharData CharSet;

void AddEngCharsToUnion(void)
{
  //  
  UCHAR ch = 'A';
  while( ch <= 'Z' )
    CharSet.SetBit(ch++);
  ch = 'a';
  while( ch <= 'z' )
    CharSet.SetBit(ch++);
}//AddEngCharsToUnion

void AddDigitsToUnion(void)
{
  UCHAR ch = '0';
  while( ch <= '9' )
    CharSet.SetBit(ch++);
}//AddDigitsToUnion

void AddSubSetToUnion(UCHAR *set)
{
  UCHAR *ch = set;

  while( *ch )
  {
    UCHAR BeginChar = *ch++;
    if( *ch=='-' && ch[1] )
    {
      UCHAR EndChar = *++ch;
      if( BeginChar <= EndChar )
        while( BeginChar <= EndChar && BeginChar )
         CharSet.SetBit(BeginChar++);
      ch++;
    }
    else
     CharSet.SetBit(BeginChar);
  }
}//AddSubSetToUnion

void InitUnion()
{
  memset( &CharSet, 0, 32 );
  AddEngCharsToUnion();
  if( adddigits )
    AddDigitsToUnion();
  AddSubSetToUnion((UCHAR *)s_AdditionalLetters);
}//InitUnion

int is_char(UCHAR c)
{
  return CharSet.GetBit(c);
}//is_char

int get_word(char *word, const char *str, int pos, int sln, int &begin_word_pos)
{
  int i,j;
  for( i = pos; i > 0 && is_char(str[i]); i-- ) ;
  if( is_char((UCHAR)str[i]) ) i--;
  begin_word_pos = i+1;

  for( j=pos; j < sln && is_char((UCHAR)str[j]); j++ ) ;
  if( is_char((UCHAR)str[j]) ) j++;
  j -= i + 1;

  if( j <= 0 )
    return( FALSE );
  //if( j > MAX_WORD_LENGTH - 1 )
  if( j > MAX_PATH - 1 )
    return( FALSE );
  strncpy(word, str+i+1, j);
  word[j] = 0;
  return( j != 0 );
}//get_word

static const int ASIZE = 256;
static UCHAR qs_bc[ ASIZE ];

void InitQuickSearch(int FKeyCode, char *word, UCHAR strlen_word)
{
  if( FKeyCode == VK_UP )
  {
    int i;
    //for( i = 0; i < ASIZE; i++ ) qs_bc[ i ] = strlen_word;
    memset(qs_bc, strlen_word, ASIZE);
    for( i = 0; i < strlen_word; i++ )
      if (qs_bc[ (UCHAR)word[ i ] ] == strlen_word)
        qs_bc[ (UCHAR)word[ i ] ] = i? (UCHAR)i: strlen_word;

    if( !casesensitive )
    {
      FSF.LStrupr(word);
      for( i = 0; i < strlen_word; i++ )
        if(qs_bc[ (UCHAR)word[ i ] ] == strlen_word )
          qs_bc[ (UCHAR)word[ i ] ] = i? (UCHAR)i: strlen_word;
    }
  }
  else //FKeyCode == VK_DOWN
  {
    int i;
    //for( i = 0; i < ASIZE; i++ ) qs_bc[ i ] = (UCHAR)(strlen_word + 1);
    memset(qs_bc, (UCHAR)(strlen_word + 1), ASIZE);
    for( i = 0; i < strlen_word; i++ )
      qs_bc[ (UCHAR)word[ i ] ] = (UCHAR)(strlen_word - i);

    if( !casesensitive )
    {
      FSF.LStrupr(word);
      for ( i = 0; i < strlen_word; i++ )
        qs_bc[ (UCHAR)word[ i ] ] = (UCHAR)(strlen_word - i);
    }
  }
}//InitQuickSearch

int QuickSearch_FW(const char *String, const char *word, int n, int m, int begin_word_pos, int SearchSelection, bool casesensitive)
{
  int i;

  i = begin_word_pos; bool res;
  while ( i <= n - m )
  {
    if( casesensitive )
      res = memcmp( &String[ i ], word, m);
    else
      res = FSF.LStrnicmp(&String[ i ], word, m);

    if( !res && (SearchSelection ||
                 (!is_char(String[i + m]) &&
                  !((i > 0) && is_char(String[i-1])))
                )
      )
      return(i);

    i += qs_bc[ (UCHAR)String[ i + m ] ];
  }
  return( -1 );
}//QuickSearch_FW

int QuickSearch_BW(const char *String, const char *word, int n, int m, int SearchSelection, bool casesensitive)
{
  int i;

  i = n - m; bool res;
  while ( i >= 0 )
  {
    if( casesensitive )
      res = memcmp( &String[ i ], word, m);
    else
      res = FSF.LStrnicmp(&String[ i ], word, m);

    if( !res && (SearchSelection ||
                 (!is_char(String[i + m]) &&
                  !((i > 0) && is_char(String[i-1]))
                 )
                )
      )
      return(i);
    i -= qs_bc[ (UCHAR)String[ i ] ];
  }
  return( -1 );
}//QuickSearch_BW

bool GetFile(char *FileName, char *&buffer, UINT &sz)
{
  buffer = NULL;
  HANDLE s = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if ( s != INVALID_HANDLE_VALUE )
  {
    sz = GetFileSize(s, NULL);
    if( sz && (buffer = new char[sz+1]) != NULL )
    {
      DWORD rb = 0;
      if ( ReadFile(s, buffer, sz, &rb, NULL) && rb )
        buffer[rb] = 0;
      else
      {
        delete buffer;
        return( FALSE );
      }
    }
  }
  CloseHandle(s);
  return( TRUE );
}//GetFile

inline int file_exists(const char *fname)
{
  return( GetFileAttributes(fname) != (DWORD)(-1) );
}//file_exists

int get_cursor_pos(int &x, int &String)
{
  struct EditorInfo EInfo;
  if( !Info.EditorControl(ECTL_GETINFO, &EInfo) )
    return( FALSE );
  x = EInfo.CurPos;
  String = EInfo.CurLine;
  return( TRUE );
}//get_cursor_pos

#ifdef _DEBUG

static char tmp_str[1500];
static char log_file_name[MAX_PATH];

void CREATE_LOG_FILE()
{
  strcpy(log_file_name, Info.ModuleName);
  int i = strlen(log_file_name) - 1;
  while( log_file_name[i] != '.' )
   i--;
  strcpy(log_file_name + i, ".log\0");
  static HANDLE HLOGFILE = CreateFile(log_file_name, GENERIC_READ|GENERIC_WRITE, 0, NULL,
    CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
  if( HLOGFILE != INVALID_HANDLE_VALUE )
    CloseHandle(HLOGFILE);
  WRITE_TO_LOG_FILE("TextNavigate activated");
}

/*
static void SysLog(char *fmt,...)
{
  va_list argptr;
  va_start(argptr, fmt);
  FILE *LogStream = fopen("C:\\BC5\\WORK\\PLUGINS\\FIND-ALL\\FIND-ALL.LOG", "at");
  vfprintf(LogStream, fmt, argptr);
  fflush(LogStream);
  fclose(LogStream);
  va_end(argptr);
}
*/
//int WRITE_TO_LOG_FILE(char *fmt,...)
int WRITE_TO_LOG_FILE(const char *str)
{
  HANDLE HF = CreateFile(log_file_name, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ, NULL,
    OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_BACKUP_SEMANTICS, NULL);
  if( HF == INVALID_HANDLE_VALUE )
    return( FALSE );

  SetFilePointer(HF, 0, NULL, FILE_END);
  char temp_str[2048];

  DWORD nchars = FarSprintf(temp_str, "%s\r\n", str);
  DWORD NumberOfBytesWritten;
  WriteFile(HF, temp_str, nchars, &NumberOfBytesWritten, NULL);
  CloseHandle(HF);
  return( TRUE );
}

#endif

void set_cursor_pos(int x, int y, struct EditorInfo *pei)
{
  struct EditorSetPosition esp;

  esp.CurLine = pei->CurLine;
  esp.CurPos = pei->CurPos;
  esp.TopScreenLine = pei->TopScreenLine;
  esp.LeftPos = pei->LeftPos;
  esp.CurTabPos = -1;
  esp.Overtype = -1;
  Info.EditorControl(ECTL_SETPOSITION,&esp);

  if( y != -1 )
  {
    esp.CurLine = y;
    esp.CurPos = x;

    esp.TopScreenLine = -1;
    esp.LeftPos = -1;
    esp.CurTabPos = -1;

    Info.EditorControl(ECTL_SETPOSITION,&esp);
  }
  Info.EditorControl(ECTL_REDRAW,NULL);
}//set_cursor_pos

int CheckForEsc(void)
{
  INPUT_RECORD rec;
  static HANDLE hConInp = GetStdHandle(STD_INPUT_HANDLE);
  DWORD ReadCount;

  PeekConsoleInput(hConInp,&rec,1,&ReadCount);
  if (ReadCount==0) return( FALSE );
  ReadConsoleInput(hConInp,&rec,1,&ReadCount);
  if (rec.EventType==KEY_EVENT)
   if (rec.Event.KeyEvent.wVirtualKeyCode==VK_ESCAPE &&
     rec.Event.KeyEvent.bKeyDown) return( TRUE );
  return( FALSE );
}//CheckForEsc

PSgmlEl GetChild(PSgmlEl parent, char *Name)
{
  if( !parent || !Name )
   return( NULL );
  PSgmlEl child = parent->child();

  if( !child )
    return( NULL );
  if( child->getname() &&
      !strcmp(child->getname(), Name) )
    return( child );
  else
    return( child->search(Name) );
}//GetChild

bool InitParam(char *&field, const char *param)
{
  if( param )
  {
    int len = strlen(param) + 1;
    field = new char[len];
    strcpy(field, param);
    return( TRUE );
  }
  return( FALSE );
}

